def picosecond2KcalMol_A_time(picosecond=1.0):
    return (20/0.978)*picosecond

def KcalMol_A_time2picosecond(KcalMol_A_time=1.0):
    return (0.978/20)*KcalMol_A_time

def KcalMol_A_force2nanonewton(KcalMol_div_A=1.0):
    return 0.06948*KcalMol_div_A

def nanonewton2KcalMol_A_force(nanonewton=1.0):
    return (1.0/0.06948)*nanonewton
